/*
 * kgraft_patch_cve-2014-3687
 *
 * Fix for CVE-2014-3687, bsc#902349
 *
 *  Copyright (c) 2015 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/kallsyms.h>
#include <net/sctp/structs.h>

static void (*kgr_sctp_chunk_hold)(struct sctp_chunk *ch);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "sctp_chunk_hold", (void *)&kgr_sctp_chunk_hold },
};

static inline bool sctp_chunk_pending(const struct sctp_chunk *chunk)
{
	return !list_empty(&chunk->list);
}

struct sctp_chunk *kgr_sctp_assoc_lookup_asconf_ack(
	const struct sctp_association *asoc, __be32 serial)
{
	struct sctp_chunk *ack;

	/* Walk through the list of cached ASCONF-ACKs and find the
	 * ack chunk whose serial number matches that of the request.
	 */
	list_for_each_entry(ack, &asoc->asconf_ack_list, transmitted_list) {
		/*
		 * Fix CVE-2014-2687
		 */
		if (sctp_chunk_pending(ack))
			continue;
		if (ack->subh.addip_hdr->serial == serial) {
			kgr_sctp_chunk_hold(ack);
			return ack;
		}
	}

	return NULL;
}

static int kgr_patch_cve_2014_3687_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}
		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int kgr_patch_cve_2014_3687_module_notify(struct notifier_block *nb,
                unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, "sctp"))
		return 0;

	ret = kgr_patch_cve_2014_3687_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_cve_2014_3687_module_nb = {
	.notifier_call = kgr_patch_cve_2014_3687_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_cve_2014_3687_init(void)
{
        int ret;

	mutex_lock(&module_mutex);
	if (find_module("sctp")) {
		ret = kgr_patch_cve_2014_3687_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_cve_2014_3687_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_cve_2014_3687_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_cve_2014_3687_module_nb);
}

